/*
 * File event notification support.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <condition_variable>
#include <list>
#include <map>
#include <mutex>

#include <fsclient/file.h>
#include <systypes/base.h>



/* File-specific notification details. */

class FileNotificationState
{
public:
    /* Synchronisation primitives for state access and notification. */

    std::mutex lock;
    std::condition_variable condition;

    /* Pending notifications for monitored files. */

    notify_flags_t pending = 0;

    /* Endpoints associated with monitored files. */

    l4_cap_idx_t endpoint = L4_INVALID_CAP;

    bool is_null() { return l4_is_invalid_cap(endpoint); }
};



/* Collection types. */

typedef std::map<file_t *, FileNotificationState> FileNotificationStates;
typedef std::map<file_t *, std::mutex> FileStateLocks;



/* An object for monitoring file event notifications. */

class FileNotifier
{
protected:
    /* General state access locking. */

    std::mutex _state_lock;

    /* File-specific state locking. */

    FileStateLocks _file_locks;

    /* Notification state. */

    FileNotificationStates _state;

    /* Notifier thread details. */

    l4_cap_idx_t _thread = L4_INVALID_CAP;
    bool _started = false;

    /* Convenience method to access file state. */

    virtual FileNotificationState &file_state(file_t *file, bool create);

    /* Helper methods. */

    virtual void _notify(file_t *file, notify_flags_t flags) = 0;

    virtual bool _transfer(FileNotificationState &state, file_t *file);

public:
    virtual ~FileNotifier();

    /* Local operations. */

    virtual long start();

    virtual long subscribe(file_t *file, notify_flags_t flags);

    virtual long unsubscribe(file_t *file);

    /* Event handling support. */

    virtual void mainloop();
};



/* An object monitoring notifications for a collection of different files. */

class GeneralFileNotifier : public FileNotifier
{
protected:
    /* Locking to protect pending notification members and to coordinate access
       to notifications. */

    std::mutex _general_lock;

    /* General lock synchronisation. */

    std::condition_variable _general_condition;

    /* Files affected by notifications. */

    std::list<file_t *> _affected;

    /* Helper methods. */

    virtual void _notify(file_t *file, notify_flags_t flags);

    virtual bool _retrieve(file_t **file);

    virtual bool _retrieve_for_file(file_t *file);

public:
    virtual long wait(file_t **file);
};



/* An object monitoring notifications for specific files. */

class SpecificFileNotifier : public FileNotifier
{
protected:
    /* Helper methods. */

    virtual void _notify(file_t *file, notify_flags_t flags);

public:
    virtual long wait_file(file_t *file);
};



/* Helper functions. */

SpecificFileNotifier *notifier_get_task_notifier();

GeneralFileNotifier *notifier_get_local_notifier();

// vim: tabstop=4 expandtab shiftwidth=4
